from collections.abc import Iterable, Iterator, MutableMapping, Sequence
from pprint import PrettyPrinter
from typing import Self

import mdformat_footnote
import mdformat_frontmatter
import mdformat_tables
import yaml
from markdown_it import MarkdownIt
from markdown_it.token import Token
from mdformat.renderer import MDRenderer
from mdit_py_plugins.footnote import footnote_plugin
from mdit_py_plugins.front_matter import front_matter_plugin

from markie.rules import render_rules, wikilink_plugin
from markie.utils import generic_repr

__all__ = ["Doc", "Section"]

MD = (
    MarkdownIt('commonmark', {'breaks': True, 'html': True})
    .use(wikilink_plugin)
    .use(front_matter_plugin)
    .use(footnote_plugin)
    .enable('table')
)

RENDERER = MDRenderer()
OPTIONS = {
    "parser_extension": [
        render_rules,
        mdformat_tables,
        mdformat_footnote,
        mdformat_frontmatter,
    ],
}

StrMap = MutableMapping[str, "str | bool | int | float | StrMap"]

HEADING_LEVELS = {"h1": 1, "h2": 2, "h3": 3, "h4": 4, "h5": 5, "h6": 6}


class Section:
    """
    A markdown section – This is mostly an ordered tree wrapper around a
    list of Tokens.
    """

    def __init__(
          self,
          heading: list[Token],
          preamble: list[Token] = None,
          subsections: list["Section"] = None,
    ):
        """
        :param heading: three markdown-it Tokens for a parsed heading tag
        :param preamble: a list of tokens between the heading title and the
            first subsection
        :param subsections: a list of this section's Sub-subsections
        """
        self.heading = heading
        self.preamble = preamble if preamble is not None else []
        self.subsections = subsections if subsections is not None else []

    @classmethod
    def from_md(cls, src: str) -> Self:
        """
        Parses the given markdown src string and returns a new Section.

        :param src: A string of markdown
        :raises ValueError: If the given markdown has multiple headings with
            the lowest heading level (e.g. two h2 headings) or if the given
            markdown contains a preamble
        """
        tokens = MD.parse(src)
        return cls.from_tokens(tokens)

    @classmethod
    def from_tokens(cls, src: Sequence[Token]) -> Self:
        """
        Creates a new Section from a sequence of markdown-it tokens

        :param src: A sequence of markdown-it tokens
        :raises ValueError: If the given markdown has multiple headings with
            the lowest heading level (e.g. two h2 headings) or if the given
            markdown contains a preamble
        """
        _, preamble, sections = parse(src)
        if len(sections) != 1 or preamble:
            raise ValueError(
                "The given markdown does not consist of a single section"
            )
        return sections[0]

    @property
    def level(self) -> int:
        """The level of this section (i.e. 1-6 corresponding to h1-h6)"""
        return HEADING_LEVELS[self.heading[0].tag]

    @property
    def title(self) -> str:
        """The raw unformatted title text of this section"""
        return self.heading[1].content

    @title.setter
    def title(self, value: str) -> None:
        """
        Sets the title of this section – the value string should be inline
        markdown.
        Be Warned — This function does not prevent the title from being
        set to non-inline text — behaviour is undefined if it is.
        """
        _, content, _ = MD.parse(value)
        self.heading[1] = content

    def prepend(self, src: "str | Section") -> None:
        """
        Prepends the given markdown string or Section *inside* this Section,
        attempting to maintain a tree-like Section structure.
        Metadata in the given src string is ignored.

        Note: this function does not prevent nonsensical operations such as
        prepending an H1 section inside an H2 section — the behaviour of this
        class is undefined in such cases.

        :param src: A string of Markdown text or a Section object
        """
        if isinstance(src, str):
            _, preamble, sections = parse(MD.parse(src))
        else:
            preamble, sections = [], [src]

        if not sections:
            self.preamble = [*preamble, *self.preamble]
        else:
            sections[-1].last().preamble.extend(self.preamble)
            self.preamble = preamble
            while (
                  self.subsections
                  and self.subsections[0].level > sections[-1].level
            ):
                current = sections[-1]
                while (
                      current.subsections
                      and self.subsections[0].level - 1 > current.level
                ):
                    current = current.subsections[-1]
                current.subsections.append(self.subsections.pop(0))
            self.subsections = [*sections, *self.subsections]

    def append(self, src: "str | Section") -> None:
        """
        Appends the given markdown string or Section to this Section,
        attempting to maintain a tree-like Section structure.
        Metadata in the given src string is ignored.

        Note: this function does not prevent nonsensical operations such as
        appending an H1 section inside an H2 section — the behaviour of this
        class is undefined in such cases.

        :param src: A string of Markdown text or a Section object
        """
        if isinstance(src, str):
            _, preamble, sections = parse(MD.parse(src))
        else:
            preamble, sections = [], [src]

        if not self.subsections:
            self.preamble.extend(preamble)
            self.subsections = sections
        else:
            self.subsections[-1].last().preamble.extend(preamble)
            while sections and self.subsections[-1].level < sections[0].level:
                current = self.subsections[-1]
                while sections[0].level - 1 > current.level:
                    current = current.subsections[-1]
                current.subsections.append(sections.pop(0))
            self.subsections = [*self.subsections, *sections]

    def last(self) -> "Section":
        """Returns the last ("rightmost") node in the section tree"""
        if self.subsections:
            return self.subsections[-1].last()
        else:
            return self

    def render(self) -> str:
        """Renders this section as markdown"""
        return RENDERER.render([*self], OPTIONS, {})

    def __iter__(self) -> Iterator[Token]:
        """
        Iterates over the tokens in this section and its subsections in order
        """
        yield from self.heading
        yield from self.preamble
        for subsection in self.subsections:
            yield from subsection

    def __repr__(self):
        return generic_repr(self)


class Doc:
    """
    A Markdown document with metadata.

    Parses Markdown text into metadata, a preamble, and Sections. Doc objects
    attempt to parse sections into a tree like-structure; however,
    this assumes markdown is "well formatted" with sections following a
    natural order (i.e. h1 -> h2 -> h3). This class does not prevent users
    from breaking this order, but its behaviour in such cases is undefined.

    Elements in the Doc preamble, and Sections consist of markdown-it Tokens
    and can be manipulated directly.
    """

    def __init__(
          self,
          metadata: StrMap,
          preamble: list[Token],
          sections: list[Section]
    ):
        """
        :param metadata: a mapping of markdown metadata
        :param preamble: a list of markdown-it Tokens that appear between the
            metadata and first heading
        :param sections: the document sections as an AST
        """
        self.metadata = metadata
        self.preamble = preamble
        self.sections = sections

    @classmethod
    def from_md(cls, src: str) -> Self:
        """
        Parses the given markdown src string and returns a new Doc.

        :param src: A string of markdown
        """
        tokens = MD.parse(src)
        return cls.from_tokens(tokens)

    @classmethod
    def from_tokens(cls, src: Sequence[Token]) -> Self:
        """
        Creates a new Doc from a sequence of markdown-it tokens

        :param src: A sequence of markdown-it tokens
        """
        frontmatter, preamble, sections = parse(src)
        return cls(frontmatter, preamble, sections)

    def prepend(self, src: str) -> None:
        """
        Prepends the given src markdown to this document, attempting to
        maintain a tree-like Section structure. Any metadata in the given src
        string is ignored

        :param src: The Markdown src string
        """
        _, preamble, sections = parse(MD.parse(src))
        if not sections:
            self.preamble = [*preamble, *self.preamble]
        else:
            sections[-1].last().preamble.extend(self.preamble)
            self.preamble = preamble
            while self.sections and self.sections[0].level > sections[-1].level:
                sections[-1].append(self.sections.pop(0))
            self.sections = [*sections, *self.sections]

    def append(self, src: str) -> None:
        """
        Appends the given Markdown src string to this document, attempting to
        maintain a tree-like Section structure. Any metadata in the given src
        string is ignored

        :param src: The Markdown src string
        """
        _, preamble, sections = parse(MD.parse(src))
        if not self.sections:
            self.preamble.extend(preamble)
            self.sections = sections
        else:
            self.sections[-1].last().preamble.extend(preamble)
            while sections and self.sections[-1].level < sections[0].level:
                self.sections[-1].append(sections.pop(0))
            self.sections = [*self.sections, *sections]

    def render(self) -> str:
        """Renders this doc as markdown"""
        if self.metadata:
            frontmatter = as_frontmatter(self.metadata)
            return RENDERER.render([frontmatter, *self], OPTIONS, {})
        else:
            return RENDERER.render(list(self), OPTIONS, {})

    def __iter__(self) -> Iterator[Token]:
        """Iterates over the Tokens of this document in order"""
        yield from self.preamble
        for section in self.sections:
            yield from section

    def __repr__(self):
        return generic_repr(self)


def parse(src: Iterable[Token]) -> tuple[StrMap, list[Token], list[Section]]:
    """
    Parses a markdown-it Token stream to metadata, a preamble, and
    nested Sections

    :param src: An Iterable of markdown-it Tokens
    :return: a tuple of the metadata, preamble, and Section tree in that order.
    """
    metadata = {}
    preamble = []
    stack = []
    tokens = iter(src)
    for token in tokens:
        if token.type == "front_matter":
            metadata = yaml.safe_load(token.content)
        elif token.type == "heading_open":
            level = HEADING_LEVELS[token.tag]
            while (len(stack) > 1
                   and level <= stack[-1].level
                   and stack[-2].level < stack[-1].level):
                stack[-2].subsections.append(stack.pop())
            heading = [token, next(tokens), next(tokens)]
            stack.append(Section(heading))
        elif stack:
            stack[-1].preamble.append(token)
        else:
            preamble.append(token)
    while len(stack) > 1 and stack[-2].level < stack[-1].level:
        stack[-2].subsections.append(stack.pop())
    return metadata, preamble, stack


def as_frontmatter(metadata: StrMap) -> Token:
    """Maps frontmatter data to a markdown-it Token"""
    return Token(
        type='front_matter',
        tag='',
        nesting=0,
        content=yaml.safe_dump(metadata, sort_keys=False),
        markup='---',
        block=True,
        hidden=True,
    )


def render(tokens: Sequence[Token]):
    """Shortcut to render with default options"""
    return RENDERER.render(tokens, OPTIONS, {})


# noinspection PyUnresolvedReferences,PyProtectedMember,PyMethodMayBeStatic
class _DocPrettyPrinter(PrettyPrinter):
    """
    Pretty printer for Doc and Section objects.

    Note: The output of this pretty printer is used in approval tests to
    check Markdown document structure.
    """

    def _pprint_doc(self, obj, stream, indent, *args):
        stream.write(f"Doc(")
        indent += 4
        stream.write(f"metadata=")
        self._format(obj.metadata, stream, indent + 9, *args)
        stream.write(f",\n{' ' * indent}preamble=")
        preamble = render(obj.preamble).strip()
        self._format(preamble, stream, indent + 9, *args)
        stream.write(f",\n{' ' * indent}sections=")
        self._format(obj.sections, stream, indent + 9, *args)
        stream.write(f")")

    def _pprint_section(self, obj, stream, indent, *args):
        stream.write(f"Section(")
        indent += 8
        stream.write(f"level=")
        self._format(obj.level, stream, indent + 6, *args)
        stream.write(f",\n{' ' * indent}title=")
        heading = obj.title.strip()
        self._format(heading, stream, indent + 6, *args)
        stream.write(f",\n{' ' * indent}preamble=")
        preamble = render(obj.preamble).strip()
        self._format(preamble, stream, indent + 9, *args)
        stream.write(f",\n{' ' * indent}subsections=")
        self._format(obj.subsections, stream, indent + 12, *args)
        stream.write(f")")

    PrettyPrinter._dispatch[Doc.__repr__] = _pprint_doc
    PrettyPrinter._dispatch[Section.__repr__] = _pprint_section



from markdown_it.rules_inline import StateInline


def wikilink_plugin(md):
    md.inline.ruler.before(
        "link",
        "wikilink",
        wikilink_rule,
    )


def wikilink_rule(state: StateInline, silent: bool) -> bool:
    start = state.pos
    pos = state.pos
    src = state.src

    if src[pos:pos+2] == "[[":
        pos += 2
    elif src[pos:pos + 3] == "![[":
        pos += 3
    else:
        return False

    while pos < state.posMax and src[pos] not in ("\n", "]"):
        pos += 1

    if src[pos:pos + 2] == "]]":
        pos += 2
    else:
        return False

    end = pos

    if not silent:
        token = state.push("wikilink", "", 0)
        token.content = src[start:end]
        # note, this is not part of markdown-it JS, but is useful for renderers
        if state.md.options.get("store_labels", False):
            token.meta["label"] = token.content

    state.pos = end
    return True



def generic_repr(obj: object):
    if hasattr(obj, "__slots__"):
        attrs = {s: getattr(obj, s) for s in obj.__slots__ if hasattr(obj, s)}
    else:
        attrs = obj.__dict__
    fields = ", ".join(f"{k}={v!r}" for k, v in attrs.items())
    return f"{type(obj).__name__}({fields})"



from pprint import pformat

import pytest
from approvaltests import verify

from markie import Section

SEC = """
# Strategies for Creative Problem Solving

The purpose of this book is to help problem solvers improve their street smarts.
"""


def test_a_single_section_can_be_parsed():
    section = Section.from_md(SEC)
    assert section.title == "Strategies for Creative Problem Solving"
    assert section.level == 1
    verify(pformat(section, width=120, sort_dicts=False))


def test_a_single_section_can_be_rendered():
    section = Section.from_md(SEC)
    verify(section.render())


def test_section_titles_can_be_changed():
    section = Section.from_md(SEC)
    assert section.title == "Strategies for Creative Problem Solving"
    section.title = "A Completely New Title!"
    assert section.title == "A Completely New Title!"
    verify(pformat(section, width=120, sort_dicts=False))


def test_section_titles_can_be_rendered_after_being_changed():
    section = Section.from_md(SEC)
    section.title = "A Completely New Title!"
    verify(section.render())


MULTIPLE_SECTIONS = """
## Heading Number 1

Content for heading number 1

## Heading Number 2

Content for heading number 2
"""


def test_multiple_sections_of_the_same_level_raise_a_value_error_when_parsed():
    with pytest.raises(ValueError):
        Section.from_md(MULTIPLE_SECTIONS)


SECTION_WITH_PREAMBLE = """
A preamble

## Heading Number 1

Content for heading number 1
"""


def test_markdown_containing_a_preamble_will_raise_a_value_error_when_parsed():
    with pytest.raises(ValueError):
        Section.from_md(SECTION_WITH_PREAMBLE)


SEC_WITH_METADATA = """
---
title: All About Envelopes
src: https://en.wikipedia.org/wiki/Windowed_envelope
---

## Envelopes

Satisfactorily-strong envelopes for business and general-purpose domestic
correspondence can be, and are, in fact, made out of paper of various qualities.
""".strip()


def test_metadata_is_ignored_when_parsing_sections():
    section = Section.from_md(SEC_WITH_METADATA)
    assert section.title == "Envelopes"
    assert section.level == 2
    verify(pformat(section, width=120, sort_dicts=False))


NESTED_SECTIONS = """
## Heading 1

Content for heading number 1

### Heading 1-1

Content for heading number 1-1

### Heading 1-2

Content for heading number 1-2
"""


def test_nested_sections_can_be_parsed():
    section = Section.from_md(NESTED_SECTIONS)
    assert len(section.preamble) != 0
    assert len(section.subsections) == 2
    assert section.title == "Heading 1"
    assert section.subsections[0].title == "Heading 1-1"
    assert section.subsections[1].title == "Heading 1-2"
    verify(pformat(section, width=120, sort_dicts=False))


def test_nested_sections_can_be_rendered():
    section = Section.from_md(NESTED_SECTIONS)
    verify(section.render())


PARA = "A completely new paragraph of text!!"


def test_a_paragraph_of_text_can_be_prepended_to_a_section():
    section = Section.from_md(SEC)
    section.prepend(PARA)
    verify(pformat(section, width=120, sort_dicts=False))


def test_a_paragraph_of_text_prepended_to_a_section_renders_correctly():
    section = Section.from_md(SEC)
    section.prepend(PARA)
    verify(section.render())


def test_a_paragraph_of_text_can_be_appended_to_a_section():
    section = Section.from_md(SEC)
    section.append(PARA)
    verify(pformat(section, width=120, sort_dicts=False))


def test_a_paragraph_of_text_appended_to_a_section_renders_correctly():
    section = Section.from_md(SEC)
    section.append(PARA)
    verify(section.render())


SEC1 = """
# An H1 Section

H1 section body
"""

SEC2 = """
## An H2 Section

H2 section body
"""


def test_an_h2_section_can_be_prepended_to_an_h1_section():
    section = Section.from_md(SEC1)
    section.prepend(SEC2)
    assert len(section.subsections) == 1
    assert section.subsections[0].title == "An H2 Section"
    verify(pformat(section, width=120, sort_dicts=False))


def test_an_h2_section_prepended_to_an_h1_section_renders_correctly():
    section = Section.from_md(SEC1)
    section.prepend(SEC2)
    verify(section.render())


def test_an_h2_section_can_be_appended_to_an_h1_section():
    section = Section.from_md(SEC1)
    section.append(SEC2)
    assert len(section.subsections) == 1
    assert section.subsections[0].title == "An H2 Section"
    verify(pformat(section, width=120, sort_dicts=False))


def test_an_h2_section_appended_to_an_h1_section_renders_correctly():
    section = Section.from_md(SEC1)
    section.append(SEC2)
    verify(section.render())


def test_an_h2_section_object_can_be_prepended_to_an_h1_section_object():
    section1 = Section.from_md(SEC1)
    section2 = Section.from_md(SEC2)
    section1.prepend(section2)
    verify(pformat(section1, width=120, sort_dicts=False))


def test_an_h2_section_object_can_be_appended_to_an_h1_section_object():
    section1 = Section.from_md(SEC1)
    section2 = Section.from_md(SEC2)
    section1.append(section2)
    verify(pformat(section1, width=120, sort_dicts=False))


SEC2_WITH_PREAMBLE = """
Some preamble before Section 2

## An H2 Section

H2 section body
"""


def test_an_h2_section_with_a_preamble_can_be_prepended_to_an_h1_section():
    section = Section.from_md(SEC1)
    section.prepend(SEC2_WITH_PREAMBLE)
    assert len(section.subsections) == 1
    assert section.subsections[0].title == "An H2 Section"
    verify(pformat(section, width=120, sort_dicts=False))


def test_an_h2_section_with_a_preamble_prepended_to_an_h1_section_renders_correctly():
    section = Section.from_md(SEC1)
    section.prepend(SEC2_WITH_PREAMBLE)
    verify(section.render())


def test_an_h2_section_with_a_preamble_can_be_appended_to_an_h1_section():
    section = Section.from_md(SEC1)
    section.append(SEC2_WITH_PREAMBLE)
    assert len(section.subsections) == 1
    assert section.subsections[0].title == "An H2 Section"
    verify(pformat(section, width=120, sort_dicts=False))


def test_an_h2_section_with_a_preamble_appended_to_an_h1_section_renders_correctly():
    section = Section.from_md(SEC1)
    section.append(SEC2_WITH_PREAMBLE)
    verify(section.render())


NESTED_SECTIONS1 = """
# An H1

h1 text

## An H2

h2 text

## Another H2

more h2 text

### An H3

h3 text
"""

NESTED_SECTIONS2 = """
## yet another h2

yet more h2 text

### Another h3

more h3 text

#### An h4

h4 text
"""

NESTED_SECTIONS3 = """
#### An h4

h4 text

### Another h3

more h3 text

## yet another h2

yet more h2 text
"""


def test_prepending_deeply_nested_sections_nest_sections_correctly():
    section = Section.from_md(NESTED_SECTIONS1)
    section.prepend(NESTED_SECTIONS2)
    verify(pformat(section, width=120, sort_dicts=False))


def test_prepended_deeply_nested_sections_render_correctly():
    section = Section.from_md(NESTED_SECTIONS1)
    section.prepend(NESTED_SECTIONS2)
    verify(section.render())


def test_appending_deeply_nested_documents_nest_sections_correctly():
    section = Section.from_md(NESTED_SECTIONS1)
    section.append(NESTED_SECTIONS3)
    verify(pformat(section, width=120, sort_dicts=False))


def test_appended_deeply_nested_documents_render_correctly():
    section = Section.from_md(NESTED_SECTIONS1)
    section.append(NESTED_SECTIONS3)
    verify(section.render())


NESTED_SECTIONS4 = """
# An H1

Some H1 Text

#### An h4

Some h4 text

## yet another h2

yet more h2 text
"""

SEC3 = """
## A new H2

Some new H2 text

### A new H3

Some new H3 text

## Another new H2

More new H2 text

### Another new H3

More new H3 text
"""


def test_prepending_a_heading_with_lower_rank_than_the_first_subheading_will_cause_the_subheading_to_nest():
    section = Section.from_md(NESTED_SECTIONS4)
    section.prepend(SEC3)
    verify(pformat(section, width=120, sort_dicts=False))



from pprint import pformat

from approvaltests import verify

from markie import Doc

PARA = "In the mind's eye one sees dinosaurs, mammoths, and sabertoothed tigers"


def test_a_paragraph_of_text_can_be_parsed():
    doc = Doc.from_md(PARA)
    assert len(doc.sections) == 0
    assert doc.metadata == {}
    verify(pformat(doc, width=120, sort_dicts=False))


def test_a_paragraph_of_text_can_be_rendered():
    doc = Doc.from_md(PARA)
    verify(doc.render())


PARA_WITH_METADATA = """
---
title: Jo Brand
src: https://en.wikipedia.org/wiki/Jo_Brand
---

Her Doc Marten boots, large size, and short hair led to false rumours 
that she was a lesbian.
""".strip()


def test_a_paragraph_of_text_with_metadata_can_be_parsed():
    doc = Doc.from_md(PARA_WITH_METADATA)
    assert len(doc.sections) == 0
    assert doc.metadata == {
        "title": "Jo Brand",
        "src": "https://en.wikipedia.org/wiki/Jo_Brand"
    }
    verify(pformat(doc, width=120, sort_dicts=False))


def test_a_paragraph_of_text_with_metadata_can_be_rendered():
    doc = Doc.from_md(PARA_WITH_METADATA)
    verify(doc.render())


def test_document_metadata_can_be_modified_and_rendered():
    doc = Doc.from_md(PARA_WITH_METADATA)
    doc.metadata["title"] += "!!!"
    doc.metadata["subtitle"] = "A biography"
    verify(doc.render())


SEC = """
# Strategies for Creative Problem Solving

The purpose of this book is to help problem solvers improve their street smarts.
"""


def test_section_can_be_parsed():
    doc = Doc.from_md(SEC)
    assert len(doc.sections) == 1
    assert len(doc.preamble) == 0
    assert doc.sections[0].title == "Strategies for Creative Problem Solving"
    assert doc.sections[0].level == 1
    assert doc.metadata == {}
    verify(pformat(doc, width=120, sort_dicts=False))


def test_section_can_be_rendered():
    doc = Doc.from_md(SEC)
    verify(doc.render())


SEC_WITH_METADATA = """
---
title: All About Envelopes
src: https://en.wikipedia.org/wiki/Windowed_envelope
---

## Envelopes

Satisfactorily-strong envelopes for business and general-purpose domestic
correspondence can be, and are, in fact, made out of paper of various qualities.
""".strip()


def test_section_with_metadata_can_be_parsed():
    doc = Doc.from_md(SEC_WITH_METADATA)
    assert doc.metadata == {
        "title": "All About Envelopes",
        "src": "https://en.wikipedia.org/wiki/Windowed_envelope"
    }
    assert len(doc.sections) == 1
    assert len(doc.preamble) == 0
    assert doc.sections[0].title == "Envelopes"
    assert doc.sections[0].level == 2
    verify(pformat(doc, width=120, sort_dicts=False))


def test_section_with_metadata_can_be_rendered():
    doc = Doc.from_md(SEC_WITH_METADATA)
    verify(doc.render())


# The Long Quiche Goodbye, Avery Aames
SEC_WITH_PRELUDE = r"""
Welcome to the grand opening of Fromagerie Bessette. Or 
as it's more commonly known by the residents of small-town Providence, 
Ohio-the Cheese Shop. Proprietor Charlotte Bessette has prepared a delightful 
sampling of bold Cabot Clothbound Cheddar, delicious tortes of Stilton and 
Mascarpone, and a taste of Sauvignon Blanc-but someone else has decided to 
make a little crime of passion the piece de resistance. Right outside the 
shop Charlotte finds a body, the victim stabbed to death with one of her 
prized olive-wood handled knives.

# Chapter 1

"I'm not dead, Charlotte," Grandpère Etienne said.\
"But you are retired, Pépère. I tweaked his rosy cheek and skirted around him 
to throw a drop cloth over the rustic wooden table that usually held wheels 
of cheese, like Abbaye de Belloc, Manchego, and Humboldt Fog, the latter 
cheese a great pairing with chardonnay. Dust billowed up as the edges of the 
drop cloth hit the shop floor.
"""


def test_section_with_preamble_can_be_parsed():
    doc = Doc.from_md(SEC_WITH_PRELUDE)
    assert len(doc.sections) == 1
    assert len(doc.preamble) != 0
    assert doc.sections[0].title == "Chapter 1"
    assert doc.sections[0].level == 1
    verify(pformat(doc, width=120, sort_dicts=False))


def test_section_with_preamble_can_be_rendered():
    doc = Doc.from_md(SEC_WITH_PRELUDE)
    verify(doc.render())


SEC_WITH_PRELUDE_AND_METADATA = """
---
title: Chromatic Fantasy
year: 1979
composer: Barbara Kolb
attributions: "text by: Howard Stern"
---

The text is divided into three parts which all interrelate.

## Part 1

The young woman at the next table is wearing a long purple dress,
and I consider inviting her to join me for lunch.
""".strip()


def test_section_with_preamble_and_metadata_can_be_parsed():
    doc = Doc.from_md(SEC_WITH_PRELUDE_AND_METADATA)
    assert doc.metadata == {
        "title": "Chromatic Fantasy",
        "year": 1979,
        "composer": "Barbara Kolb",
        "attributions": "text by: Howard Stern",
    }
    assert len(doc.sections) == 1
    assert len(doc.preamble) != 0
    assert doc.sections[0].title == "Part 1"
    assert doc.sections[0].level == 2
    verify(pformat(doc, width=120, sort_dicts=False))


def test_section_with_preamble_and_metadata_can_be_rendered():
    doc = Doc.from_md(SEC_WITH_PRELUDE_AND_METADATA)
    verify(doc.render())


MULTIPLE_SECTIONS = """
## Heading Number 1

Content for heading number 1

## Heading Number 2

Content for heading number 2
"""


def test_multiple_sections_of_the_same_level_can_be_parsed():
    doc = Doc.from_md(MULTIPLE_SECTIONS)
    assert len(doc.sections) == 2
    assert len(doc.preamble) == 0
    assert doc.sections[0].title == "Heading Number 1"
    assert doc.sections[0].level == 2
    assert doc.sections[1].title == "Heading Number 2"
    assert doc.sections[1].level == 2
    verify(pformat(doc, width=120, sort_dicts=False))


def test_multiple_sections_of_the_same_level_can_be_render():
    doc = Doc.from_md(MULTIPLE_SECTIONS)
    verify(doc.render())


NESTED_SECTIONS = """
## Heading Number 1

Content for heading number 1

### Heading 1-1

Content for heading number 1-1

### Heading 1-2

Content for heading number 1-2

## Heading Number 2

Content for heading number 2

### Heading 2-1

Content for heading number 2-1
"""


def test_nested_sections_can_be_parsed():
    doc = Doc.from_md(NESTED_SECTIONS)
    assert len(doc.sections) == 2
    assert len(doc.preamble) == 0
    assert doc.sections[0].title == "Heading Number 1"
    assert doc.sections[0].level == 2
    assert doc.sections[1].title == "Heading Number 2"
    assert doc.sections[1].level == 2
    verify(pformat(doc, width=120, sort_dicts=False))


def test_nested_sections_can_be_rendered():
    doc = Doc.from_md(NESTED_SECTIONS)
    verify(doc.render())


PARA1 = "The first line of text"
PARA2 = "The second line of text"


def test_two_paragraphs_can_be_prepended():
    doc = Doc.from_md(PARA2)
    doc.prepend(PARA1)
    assert len(doc.preamble) != 0
    assert len(doc.sections) == 0
    assert doc.metadata == {}
    verify(pformat(doc, width=120, sort_dicts=False))


def test_two_prepended_paragraphs_are_rendered_correctly():
    doc = Doc.from_md(PARA2)
    doc.prepend(PARA1)
    verify(doc.render())


def test_two_paragraphs_can_be_appended():
    doc = Doc.from_md(PARA1)
    doc.append(PARA2)
    assert len(doc.preamble) != 0
    assert len(doc.sections) == 0
    assert doc.metadata == {}
    verify(pformat(doc, width=120, sort_dicts=False))


def test_two_appended_paragraphs_are_rendered_correctly():
    doc = Doc.from_md(PARA1)
    doc.append(PARA2)
    verify(doc.render())


PARA_WITH_METADATA1 = """
---
title: The main paragraph
---

The original text
""".strip()

PARA_WITH_METADATA2 = """
---
title: The second paragraph
field1: an ignored field
---

The appended/prepended text
""".strip()


def test_two_paragraphs_with_metadata_can_be_prepended():
    doc = Doc.from_md(PARA_WITH_METADATA1)
    doc.prepend(PARA_WITH_METADATA2)
    assert len(doc.preamble) != 0
    assert len(doc.sections) == 0
    assert doc.metadata == {"title": "The main paragraph"}
    verify(pformat(doc, width=120, sort_dicts=False))


def test_two_prepended_paragraphs_with_metadata_are_rendered_correctly():
    doc = Doc.from_md(PARA_WITH_METADATA1)
    doc.prepend(PARA_WITH_METADATA2)
    verify(doc.render())


def test_two_paragraphs_with_metadata_can_be_appended():
    doc = Doc.from_md(PARA_WITH_METADATA1)
    doc.append(PARA_WITH_METADATA2)
    assert len(doc.preamble) != 0
    assert len(doc.sections) == 0
    assert doc.metadata == {"title": "The main paragraph"}
    verify(pformat(doc, width=120, sort_dicts=False))


def test_two_appended_paragraphs_with_metadata_are_rendered_correctly():
    doc = Doc.from_md(PARA_WITH_METADATA1)
    doc.append(PARA_WITH_METADATA2)
    verify(doc.render())


SEC1 = """
# First H1

First H1 section
"""

SEC2 = """
# Second H1

Second H1 section
"""


def test_two_sections_of_the_same_level_can_be_prepended():
    doc = Doc.from_md(SEC2)
    doc.prepend(SEC1)
    assert len(doc.sections) == 2
    assert len(doc.preamble) == 0
    assert doc.metadata == {}
    assert doc.sections[0].title == "First H1"
    assert doc.sections[0].level == 1
    assert doc.sections[1].title == "Second H1"
    assert doc.sections[1].level == 1
    verify(pformat(doc, width=120, sort_dicts=False))


def test_two_prepended_sections_of_the_same_level_can_be_rendered():
    doc = Doc.from_md(SEC2)
    doc.prepend(SEC1)
    verify(doc.render())


def test_two_sections_of_the_same_level_can_be_appended():
    doc = Doc.from_md(SEC1)
    doc.append(SEC2)
    assert len(doc.sections) == 2
    assert len(doc.preamble) == 0
    assert doc.metadata == {}
    assert doc.sections[0].title == "First H1"
    assert doc.sections[0].level == 1
    assert doc.sections[1].title == "Second H1"
    assert doc.sections[1].level == 1
    verify(pformat(doc, width=120, sort_dicts=False))


def test_two_appended_sections_of_the_same_level_can_be_rendered():
    doc = Doc.from_md(SEC1)
    doc.append(SEC2)
    verify(doc.render())


SEC1_WITH_PREAMBLE = """
The first preamble

# First H1

First H1 section
"""

SEC2_WITH_PREAMBLE = """
The second preamble

# Second H1

Second H1 section
"""


def test_two_sections_of_the_same_level_with_preambles_can_be_prepended():
    doc = Doc.from_md(SEC2_WITH_PREAMBLE)
    doc.prepend(SEC1_WITH_PREAMBLE)
    assert len(doc.sections) == 2
    assert len(doc.preamble) != 0
    assert doc.metadata == {}
    assert doc.sections[0].title == "First H1"
    assert doc.sections[0].level == 1
    assert doc.sections[1].title == "Second H1"
    assert doc.sections[1].level == 1
    verify(pformat(doc, width=120, sort_dicts=False))


# noinspection PyPep8
def test_two_prepended_sections_of_the_same_level_with_preambles_can_be_rendered():
    doc = Doc.from_md(SEC2_WITH_PREAMBLE)
    doc.prepend(SEC1_WITH_PREAMBLE)
    verify(doc.render())


def test_two_sections_of_the_same_level_with_preambles_can_be_appended():
    doc = Doc.from_md(SEC1_WITH_PREAMBLE)
    doc.append(SEC2_WITH_PREAMBLE)
    assert len(doc.sections) == 2
    assert len(doc.preamble) != 0
    assert doc.metadata == {}
    assert doc.sections[0].title == "First H1"
    assert doc.sections[0].level == 1
    assert doc.sections[1].title == "Second H1"
    assert doc.sections[1].level == 1
    verify(pformat(doc, width=120, sort_dicts=False))


# noinspection PyPep8
def test_two_appended_sections_of_the_same_level_with_preambles_can_be_rendered():
    doc = Doc.from_md(SEC1_WITH_PREAMBLE)
    doc.append(SEC2_WITH_PREAMBLE)
    verify(doc.render())


AN_H1 = """
# An H1 Heading

Some H1 Heading text
"""

AN_H2 = """
## An H2 Heading

Some H2 Heading text
"""


# noinspection PyPep8
def test_prepending_a_lower_level_heading_to_a_higher_level_heading_will_cause_the_higher_level_heading_to_nest():
    doc = Doc.from_md(AN_H2)
    doc.prepend(AN_H1)
    assert len(doc.sections) == 1
    assert doc.sections[0].title == "An H1 Heading"
    assert len(doc.sections[0].subsections) == 1
    assert doc.sections[0].subsections[0].title == "An H2 Heading"
    verify(pformat(doc, width=120, sort_dicts=False))


def test_prepending_a_higher_level_heading_to_a_lower_level_heading_will_NOT_cause_the_higher_level_heading_to_nest():
    doc = Doc.from_md(AN_H1)
    doc.prepend(AN_H2)
    assert len(doc.sections) == 2
    assert doc.sections[0].title == "An H2 Heading"
    assert doc.sections[1].title == "An H1 Heading"
    verify(pformat(doc, width=120, sort_dicts=False))


def test_prepended_nested_sections_can_be_rendered():
    doc = Doc.from_md(AN_H2)
    doc.prepend(AN_H1)
    verify(doc.render())


def test_prepended_sections_with_different_levels_can_be_rendered():
    doc = Doc.from_md(AN_H1)
    doc.prepend(AN_H2)
    verify(doc.render())


def test_appending_a_higher_level_heading_to_a_lower_level_heading_will_cause_the_higher_level_heading_to_nest():
    doc = Doc.from_md(AN_H1)
    doc.append(AN_H2)
    assert len(doc.sections) == 1
    assert doc.sections[0].title == "An H1 Heading"
    assert len(doc.sections[0].subsections) == 1
    assert doc.sections[0].subsections[0].title == "An H2 Heading"
    verify(pformat(doc, width=120, sort_dicts=False))


def test_appending_a_lower_level_heading_to_a_higher_level_heading_will_NOT_cause_the_lower_level_heading_to_nest():
    doc = Doc.from_md(AN_H2)
    doc.append(AN_H1)
    assert len(doc.sections) == 2
    assert doc.sections[0].title == "An H2 Heading"
    assert doc.sections[1].title == "An H1 Heading"
    verify(pformat(doc, width=120, sort_dicts=False))


def test_appended_nested_sections_can_be_rendered():
    doc = Doc.from_md(AN_H1)
    doc.append(AN_H2)
    verify(doc.render())


def test_appended_sections_with_different_levels_can_be_rendered():
    doc = Doc.from_md(AN_H2)
    doc.append(AN_H1)
    verify(doc.render())


NESTED_SECTIONS1 = """
# An H1

h1 text

## An H2

h2 text

## Another H2

more h2 text

### An H3

h3 text
"""

NESTED_SECTIONS2 = """
#### An h4

h4 text

### Another h3

more h3 text

## yet another h2

yet more h2 text

# Another h1?

For testing purposes
"""


def test_prepending_deeply_nested_documents_nest_sections_correctly():
    doc = Doc.from_md(NESTED_SECTIONS2)
    doc.prepend(NESTED_SECTIONS1)
    verify(pformat(doc, width=120, sort_dicts=False))


def test_prepended_deeply_nested_documents_render_correctly():
    doc = Doc.from_md(NESTED_SECTIONS2)
    doc.prepend(NESTED_SECTIONS1)
    verify(doc.render())


def test_appending_deeply_nested_documents_nest_sections_correctly():
    doc = Doc.from_md(NESTED_SECTIONS1)
    doc.append(NESTED_SECTIONS2)
    verify(pformat(doc, width=120, sort_dicts=False))


def test_appended_deeply_nested_documents_render_correctly():
    doc = Doc.from_md(NESTED_SECTIONS1)
    doc.append(NESTED_SECTIONS2)
    verify(doc.render())


def test_preambles_will_nest_into_the_rightmost_leaf_section_when_prepended_to():
    doc = Doc.from_md("The original preamble")
    doc.prepend(NESTED_SECTIONS1)
    assert len(doc.preamble) == 0
    verify(pformat(doc, width=120, sort_dicts=False))


def test_preambles_will_nest_into_the_rightmost_leaf_section_when_appended():
    doc = Doc.from_md(NESTED_SECTIONS1)
    doc.append("A new preamble")
    assert len(doc.preamble) == 0
    verify(pformat(doc, width=120, sort_dicts=False))
